/*****************************************************************************
*   $Id$
*
*   Copyright (c) 1996-1998, Darren Hiebert
*
*   This source code is released for free distribution under the terms of the
*   GNU General Public License.
*
*   This module contains functions supporting resizeable strings.
*****************************************************************************/

/*============================================================================
=   Include files
============================================================================*/
#include "general.h"

#if defined(HAVE_STDLIB_H)
# include <stdlib.h>	/* to declare malloc() */
#endif
#include <limits.h>	/* to define INT_MAX */

#define VSTRING_WRITE
#include "vstring.h"

#include "debug.h"
#include "main.h"

/*============================================================================
=   Data definitions
============================================================================*/
const size_t vStringInitialSize = 32;

/*============================================================================
=   Function prototypes
============================================================================*/
static void vStringInit __ARGS((vString *const string));
static void vStringAlloc __ARGS((vString *const string, const size_t size));
static boolean vStringResize __ARGS((vString *const string, const size_t newSize));

/*============================================================================
=   Function definitions
============================================================================*/

static void vStringInit( string )
    vString *const string;
{
    string->length	= 0;
    string->size	= 0;
    string->buffer	= NULL;
}

static void vStringAlloc( string, size )
    vString *const string;
    const size_t size;
{
    string->length = 0;
    string->size = size;
    string->buffer = (char *)malloc(string->size);
    if (string->buffer == NULL)
	error(FATAL | PERROR, "out of memory");
}

static boolean vStringResize( string, newSize )
    vString *const string;
    const size_t newSize;
{
    boolean ok = FALSE;
    char *const newBuffer = (char *)realloc(string->buffer, newSize);

    if (newBuffer != NULL)
    {
	string->size = newSize;
	string->buffer = newBuffer;
	ok = TRUE;
    }
    return ok;
}

/*----------------------------------------------------------------------------
*-	External interface
----------------------------------------------------------------------------*/

extern boolean vStringAutoResize( string )
    vString *const string;
{
    boolean ok = TRUE;

    if (string->size <= INT_MAX / 2)
    {
	const size_t newSize = string->size * 2;

	if (! vStringResize(string, newSize))
	    error(FATAL | PERROR, "input token too big; out of memory");
    }
    return ok;
}

extern void vStringClear( string )
    vString *const string;
{
    string->length = 0;
    string->buffer[0] = '\0';
    DebugStatement( clearString(string->buffer, string->size); )
}

extern void vStringDelete( string )
    vString *const string;
{
    if (string->buffer != NULL)
	free(string->buffer);
    free(string);
}

extern vString *vStringNew()
{
    vString *string = (vString *)malloc(sizeof(vString));

    if (string == NULL)
	error(FATAL | PERROR, "out of memory");
    else
    {
	vStringInit(string);
	vStringAlloc(string, vStringInitialSize);
    }
    return string;
}

extern void vStringPut( string, c )
    vString *const string;
    const int c;
{
    if (string->length == string->size)		/*  check for buffer overflow */
	vStringAutoResize(string);

    string->buffer[string->length] = c;
    if (c != '\0')
	string->length++;
}

extern void vStringCat( string, s )
    vString *const string;
    const char *const s;
{
    const char *p = s;

    do
	vStringPut(string, *p);
    while (*p++ != '\0');
}
